program inst_eddies

  ! The current directory at run-time should contain directory SHPC.

  use, intrinsic:: ieee_arithmetic, only: IEEE_SUPPORT_DATATYPE, &
       ieee_support_nan
  use, intrinsic:: ISO_FORTRAN_ENV, only: ERROR_UNIT

  ! Libraries:
  use jumble, only: assert, new_unit, argwhere, get_command_arg_dyn
  use numer_rec_95, only: indexx

  use config_m, only: config, max_radius
  use derived_types, only: eddy, shpc_slice_handler
  use input_ssh_m, only: input_ssh, nlon, nlat
  use nearby_extr_m, only: nearby_extr
  use set_all_extr_m, only: set_all_extr
  use set_contours_m, only: set_contours
  use shpc_close_m, only: shpc_close
  use shpc_create_m, only: shpc_create
  use shpc_open_m, only: shpc_open
  use write_snapshot_m, only: write_snapshot

  implicit none

  type(eddy), allocatable:: list(:) ! (number_extr) "Eddies" at a
  ! given date. In general, there may be both cyclonic and
  ! anticyclonic eddies in the list. Eddies include "eddies"
  ! without an outermost contour, that is, only the extremum is
  ! defined. The subscript value in list is the "identification
  ! number" of the extremum.

  integer number_extr ! number of extrema

  integer, allocatable:: extr_map(:, :)
  ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is
  ! periodic in longitude, else (nlon, nlat). At a point of extremum
  ! SSH: identification number or this extremum. 0 at other points.

  TYPE(shpc_slice_handler) hshpc_cyclo, hshpc_anti
  integer iostat, i, l, n_arg
  character(len = 200) iomsg
  integer unit

  real step(2) ! longitude and latitude steps, in rad
  logical periodic ! grid is periodic in longitude

  real, allocatable:: ssh(:, :)
  ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is periodic
  ! in longitude, else (nlon, nlat). Sea-surface height, in m.

  real, allocatable:: u(:, :), v(:, :)
  ! (1 - max_radius(1):nlon + max_radius(1), nlat) if periodic, else
  ! (nlon, nlat). Wind, in m s-1.

  logical exist
  real t0, t1 ! CPU times, in s
  character(len = 30) timings_fname ! file name
  integer:: date = 0, slice = 0
  namelist /dates_nml/ date, slice

  integer, allocatable:: sorted_extr(:) ! (number_extr)
  ! Sorted identifying number of extrema: first, minima sorted in
  ! order of decreasing SSH, and second, maxima sorted in order of
  ! increasing SSH.

  integer, allocatable:: selection(:)
  ! identifying numbers of a selection of extrema

  integer n_cycl ! number of cyclones
  integer llc(2) ! indices in global grid of lower left corner
  integer urc(2) ! indices in global grid of upper right corner

  integer, allocatable:: outside_points(:, :) ! (2, :)
  ! coordinates in projection space of all the significant extrema,
  ! except the target extremum

  character(len = :), allocatable:: shpc_dir
  character(len = :), allocatable:: ssh_fname, u_fname, v_fname ! file names

  !--------------------------------------------------------------

  call cpu_time(t0)
  call assert(IEEE_SUPPORT_DATATYPE(0.), ieee_support_nan(0.), &
       "inst_eddies: not enough IEEE support")
  n_arg = command_argument_count()
  call assert(n_arg >= 2 .and. n_arg <= 4, &
       "Usage: SHPC-directory inst_eddies h-file [u-file [v-file]]")
  call get_command_arg_dyn(1, shpc_dir)
  call get_command_arg_dyn(2, ssh_fname)

  if (n_arg == 2) then
     u_fname = ssh_fname
     v_fname = ssh_fname
  else
     call get_command_arg_dyn(3, u_fname)

     if (n_arg == 3) then
        v_fname = u_fname
     else
        call get_command_arg_dyn(4, v_fname)
     end if
  end if

  print *, "inst_eddies: Enter dates_nml:"
  read(unit = *, nml = dates_nml)
  write(timings_fname, fmt = '(a, "/Slice_", i0, "/timings.txt")') shpc_dir, &
       slice
  inquire(file = timings_fname, exist = exist)
  call new_unit(unit)

  if (exist) then
     open(unit, file = timings_fname, STATUS = "old", ACTION = "WRITE", &
          position = "append")
  else
     OPEN(unit, file = timings_fname, STATUS = "new", ACTION = "WRITE", &
          iostat = iostat, iomsg = iomsg)

     if (iostat /= 0) then
        write(ERROR_UNIT, fmt = "(a, a, i0, a)") "Is the directory ", &
             shpc_dir, "/Slice_", slice, " created?"
        write(ERROR_UNIT, fmt = *) trim(iomsg)
        stop 1
     end if
  end if

  call config
  call input_ssh(step, periodic, ssh, u, v, ssh_fname, u_fname, v_fname)

  ! Open or create now so we test the existence of directories. Having
  ! called config, we know that directory SHPC_dir exists, but we do
  ! not know about the sub-directories of SHPC_dir.
  call shpc_open(hshpc_cyclo, shpc_dir, cyclone = .true., slice = slice, &
       pszaccess = "rb+", iostat = iostat)

  if (iostat == 0) then
     call shpc_open(hshpc_anti, shpc_dir, cyclone = .false., slice = slice, &
          pszaccess = "rb+")
  else
     print *, "inst_eddies: Opening failed so we will create a slice..."
     call shpc_create(hshpc_cyclo, shpc_dir, cyclone = .true., slice = slice, &
          grid_lon_lat = .true.)
     call shpc_create(hshpc_anti, shpc_dir, cyclone = .false., slice = slice, &
          grid_lon_lat = .true.)
  end if

  call cpu_time(t1)
  write(unit, fmt = *) "CPU time for preamble, before computation:", t1 - t0, &
       "s"
  t0 = t1
  call set_all_extr(extr_map, list, step, periodic, ssh)
  number_extr = size(list)
  allocate(sorted_extr(number_extr))

  forall (i = 1:number_extr)
     list(i)%out_cont%closed = .true.
     ! must be intialized to true because it is used in nearby_extr
  end forall

  ! Double sort of extrema on cyclonicity and SSH value:

  selection = argwhere(list%cyclone)
  n_cycl = size(selection)
  selection = selection(indexx(list(selection)%extr%ssh))
  sorted_extr(:n_cycl) = selection(n_cycl:1:- 1) ! descending order of ssh

  selection = argwhere(.not. list%cyclone)
  selection = selection(indexx(list(selection)%extr%ssh))
  sorted_extr(n_cycl + 1:) = selection

  ! Done sorting

  do l = 1, number_extr
     i = sorted_extr(l)

     ! Define the geographical window around each eddy extremum:

     llc = list(i)%extr%coord_proj - max_radius
     urc = list(i)%extr%coord_proj + max_radius

     llc(2) = max(llc(2), 1)
     urc(2) = min(urc(2), nlat)

     if (.not. periodic) then
        llc(1) = max(llc(1), 1)
        urc(1) = min(urc(1), nlon)
     end if

     ! Done defining geographical window

     outside_points &
          = nearby_extr(extr_map(llc(1):urc(1), llc(2):urc(2)), list, i)

     ! Shift the x coordinate of each point to a value close to the
     ! coordinate of the target extremum:
     if (periodic) outside_points(1, :) = outside_points(1, :) &
          + ceiling((llc(1) - outside_points(1, :)) / real(nlon)) * nlon

     call set_contours(list(i)%out_cont, list(i)%speed_cont, &
          list(i)%max_speed, list(i)%cyclone, list(i)%extr, &
          list(i)%innermost_level, step, ssh(llc(1):urc(1), llc(2):urc(2)), &
          u(llc(1):urc(1), llc(2):urc(2)), v(llc(1):urc(1), llc(2):urc(2)), &
          llc, outside_points)
  end do

  call cpu_time(t1)
  write(unit, fmt = *) "CPU time for computation, before output:", t1 - t0, "s"
  t0 = t1
  print *, "inst_eddies: number_extr = ", number_extr
  call write_snapshot(list, hshpc_cyclo, hshpc_anti, date)
  CALL shpc_close(hshpc_cyclo)
  CALL shpc_close(hshpc_anti)

  if (iostat == 0) then
     print *, 'inst_eddies: Appended to shapefiles in one slice of ', &
          SHPC_dir, '.'
  else
     print *, 'inst_eddies: Created shapefiles in one slice of ', SHPC_dir, '.'
  end if

  call cpu_time(t1)
  write(unit, fmt = *) "CPU time for output:", t1 - t0, "s"
  close(unit)

end program inst_eddies
